/*
* VitalityParser mxy object for Max/MSP
* reads in data file for icad2006 concert and retrives data
* 
* Danny Shapira 4_20_2006
*
*/

import com.cycling74.max.*;

import java.io.*;
import java.util.ArrayList;
import java.util.Iterator;



public class VitalityParser extends MaxObject {

		public static final int FIELDS = 46; 		//number of fields that the data contains
		public static final int ROWS = 300; 		//number of rows
		double latSkew = 3;	//allowable error for lat searching
		double lonSkew = 3;	//alloeable error for lon searching 
		
		int numCountriesToGet = 1;
		
		ArrayList dataList = new ArrayList();
		ArrayList dataListNormalized = new ArrayList();
		ArrayList countryList = new ArrayList();
		ArrayList categoryList = new ArrayList();
	
		public final String [] categories = {"Country","Region","Area_km2","Population",	"GDP_dollar",
				"GDP_per_capita","Capital","Latitude",	"Longitude",	"Latitude_deg","Longitude_deg",
				"Sanitation_total","Sanitation_urban",	"Sanitation_rural","Drinking_total",
				"Drinking_urban","Drinking_rural","Life_Expectancy_Male","Life_Expectancy_Female",
				"School_Total","School_Male",	"School_Female", "Housing_total",	"Housing_urban",
				"Housing_rural","Unemployment_both","Unemployment_male","Unemployment_female",
				"Marriage_male",	"Marriage_female","Contraceptives","Fertility","Youth_fertility",	
				"HIV_AIDS", "HIV_rate_women","Illiteracy_male","Illiteracy_female","Employees_male",
				"Employees_female","Employers_male","Employers_female",	"Own_account_male",
				"Own_account_female","Family_worker_male",	"Family_worker_female","Density" };
		
		public final String [] countries = {
		"Afghanistan", "Albania", "Algeria", "Andorra", "Angola", 
		"Antigua_and_Barbuda", "Argentina", "Armenia", "Australia", "Austria", "Azerbaijan", "Bahamas", 
		"Bahrain", "Bangladesh", "Barbados", "Belarus", "Belgium", "Belize", "Benin", "Bhutan", "Bolivia", 
		"Bosnia_and_Herzegovina", "Botswana", "Brazil", "Brunei", "Bulgaria", "Burkina_Faso", "Burma", "Burundi", 
		"Cambodia", "Cameroon", "Canada", "Cape_Verde", "Central_African_Republic", "Chad", "Chile", "China", 
		"Colombia", "Comoros", "Congo,_Democratic_Republic", "Congo,_Republic", "Costa_Rica", "Cote_d'Ivoire", 
		"Croatia", "Cuba", "Cyprus", "Czech_Republic", "Denmark", "Djibouti", 
		"Dominica", "Dominican_Republic", "East_Timor", "Ecuador", "Egypt", "El_Salvador", "Equatorial_Guinea", 
		"Eritrea", "Estonia", "Ethiopia", "Fiji", "Finland", "France", "Gabon", "Gambia", "Georgia", "Germany", 
		"Ghana", "Greece", "Grenada", "Guatemala", "Guinea", "Guyana", "Haiti", "Honduras", "Hong_Kong", "Hungary", 
		"Iceland", "India", "Indonesia", "Iran", "Iraq", "Ireland", "Israel", "Italy", "Jamaica", "Japan", 
		"Jordan", "Kazakhstan", "Kenya", "Kiribati", "Korea,_North", "Korea,_South", "Kuwait", "Kyrgyzstan", 
		"Laos", "Latvia", "Lebanon", "Lesotho", "Liberia", "Libya", "Liechtenstein", "Lithuania", "Luxembourg", 
		"Macedonia", "Madagascar", "Malawi", "Malaysia", "Maldives", "Mali", "Malta", "Mauritania", 
		"Mauritius", "Mayotte", "Mexico", "Moldova", "Monaco", "Mongolia", "Morocco", "Mozambique", "Namibia", 
		"Nepal", "Netherlands", "New_Zealand", "Nicaragua", "Niger", "Nigeria", "Norway", "Oman", "Pakistan", 
		"Palau", "Palestina", "Panama", "Papua_New_Guinea", "Paraguay", "Peru", "Philippines", "Poland", "Portugal", 
		"Puerto_Rico", "Qatar", "Reunion", "Romania", "Russia", "Rwanda", "Saint_Helena", "Samoa", "San_Marino", 
		"Saudi_Arabia", "Senegal", "Serbia_and_Montenegro", "Seychelles", "Sierra_Leone", "Singapore", "Slovakia", 
		"Slovenia", "Solomon_Islands", "Somalia", "South_Africa", "Spain", "Sri_Lanka", "Sudan", "Suriname", 
		"Swaziland", "Sweden", "Switzerland", "Syria", "Taiwan", "Tajikistan", "Tanzania", "Thailand", 
		"Togo", "Tonga", "Trinidad_and_Tobago", "Tunisia", "Turkey", "Turkmenistan", "Tuvalu", "Uganda", 
		"Ukraine", "United_Arab_Emirates", "United_Kingdom", "United_States", "Uruguay", "Uzbekistan", 
		"Vanuatu", "Venezuela", "Vietnam", "Yemen", "Zambia", "Zimbabwe" };
		
		
		
		VitalityParser(Atom[] a ) {
			declareIO(2, 2);
		}
		
		public void loadbang() {
			init();
		}
		
        public void bang() {
                post("hello bang!");
                outletBang(0);
        }
        
        public void init() {
	        	for(int i=0; i<categories.length; i++){
	    			categoryList.add(categories[i]);
	    		}
	        	for(int i=0; i<countries.length; i++){
	        		countryList.add(countries[i]);
        		}
	        	
	        	post("VitalityParser version 0.11 written by Danny Shapira 5_04_06");
        }
        
        //***** read in formatted data and fill arrayLists*****//
        public void read(Atom[] args) throws IOException {
	        	Atom[] aLoad = new Atom[2];
	        	aLoad[0] = Atom.newAtom("doneLoading");
	        	aLoad[1] = Atom.newAtom(0);
	        	outlet(1, aLoad);
        	
        		init();
        		FileReader inputFile = new FileReader(Atom.toOneString(args));
	        	BufferedReader in = new BufferedReader(inputFile);

	        	ArrayList cList = new ArrayList();
	        	String tmp;
	        	String s1 = new String();
	        	String s2 = new String();
	        	boolean end = false;
	        	int lineNum = 1;
	        	int elementNum = 0;
  	
	        	while ((tmp = in.readLine()) != null ) {
	        		end = false;
	        		elementNum = 0;
	        		cList = new ArrayList();
	        		
	        		while(!end && lineNum != 1) {
	        			
	        			int i = tmp.indexOf("\t");  //search for a n instance of tabulation
	        			
	        			if(i != -1) {
	        				s1 = tmp.substring(0, i); //store into s1 the string contained in the cell
	        				s2 = tmp.substring(i+1, tmp.length());
	        				tmp = s2; //store into tmp the rest of the line
	        			}
	        			else if(i == -1) {
	        				s1 = tmp;
	        				end = true;
	        			}	
	        			
	        			cList.add(s1);
	        			//post("" + (String)countryList.get(elementNum));
	        			elementNum++;
	        		}
	        		if(cList.size() > 0) 
	        			dataList.add(cList);
	        		lineNum++;
	        	}
	        	in.close();
	        	
	        	// run other setup routines //
	        // outputCountryMenu();
	        // outputCellblockData();       	
	        	aLoad = new Atom[2];
	        	aLoad[0] = Atom.newAtom("doneLoading");
	        	aLoad[1] = Atom.newAtom(1);
    			outlet(1, aLoad);
        	}
        
        //***** output data in jit.cellblock format*****//
        public void outputCellblockData() {
        	Thread t = new Thread(new Runnable(){
        		public void run()
        		{		
        			Atom[] aClear = new Atom[3];
        			aClear[0] = Atom.newAtom("cellblockData");
        			aClear[1] = Atom.newAtom("clear");
        			aClear[2] = Atom.newAtom("all");
        			outlet(1, aClear);
          		
        			//populate the top category row //
            		Atom[] a = new Atom[5];
            		for(int j=0; j<categories.length; j++) {
            			a[0] = Atom.newAtom(new String("cellblockData")); 
      				a[1] = Atom.newAtom(new String("set"));
      				a[2] = Atom.newAtom(j);
      				a[3] = Atom.newAtom(0);
      				a[4] = Atom.newAtom(categories[j]); 
      				outlet(1, a ); 
            		}
            		
        			//fill in all the country data //
        			Iterator iter = dataList.iterator(); 
            		ArrayList e = null;
            		int i = 1; //don't overwrite the country data //

            		while(iter.hasNext()) {	
            			e = (ArrayList)iter.next();
          			for(int j=0; j<e.size(); j++) { 
          				a[0] = Atom.newAtom(new String("cellblockData")); 
          				a[1] = Atom.newAtom(new String("set"));
          				a[2] = Atom.newAtom(j);
          				a[3] = Atom.newAtom(i);
          				a[4] = Atom.newAtom((String)e.get(j)); 
          				outlet(1, a ); 
          			}         			
          			i++;
            		}
        		}	
        	});
        	t.start(); 
        }
        	
        //***** output country menu  *****//
        public void outputCountryMenu() {
        	Thread t = new Thread(new Runnable(){
        		public void run()
        		{		
        			Atom[] aClear = new Atom[2];
        			aClear[0] = Atom.newAtom("countryMenu");
        			aClear[1] = Atom.newAtom("clear");
        			outlet(1, aClear);
        			
        			Atom[] a = new Atom[3];
        			
        			for(int i=0; i<countries.length; i++) { 
        				a[0] = Atom.newAtom(new String("countryMenu")); 
            			a[1] = Atom.newAtom(new String("append"));
        				a[2] = Atom.newAtom(countries[i]); 
        				outlet(1, a ); 
        			}	
        			
        		}	
        	});
        	t.start();
        }
        
        //***** output category menu  *****//
        public void outputCategoryMenu() {
        	Thread t = new Thread(new Runnable(){
        		public void run()
        		{		
        			Atom[] aClear = new Atom[2];
        			aClear[0] = Atom.newAtom("categoryMenu");
        			aClear[1] = Atom.newAtom("clear");
        			outlet(1, aClear);
        			
        			Atom[] a = new Atom[3];
        			
        			for(int i=0; i<categories.length; i++) { 
        				a[0] = Atom.newAtom(new String("categoryMenu")); 
            			a[1] = Atom.newAtom(new String("append"));
        				a[2] = Atom.newAtom(categories[i]); 
        				outlet(1, a ); 
        			}	
        			
        		}	
        	});
        	t.start();
        }
        
        //***** find closest country match given input lat and lon *****//
        public void latLonSearch(double latIn, double lonIn) {
        		int latIndex = categoryList.indexOf("Latitude");
        		int lonIndex = categoryList.indexOf("Longitude");
        		//clear the messageBox//
        		Atom[] aClear = new Atom[2];
			aClear[0] = Atom.newAtom("latLonSearchResults");
			aClear[1] = Atom.newAtom("set");
			outlet(2, aClear);

        		boolean foundMatch = false;
        		int i = 1;
        		int foundCountries = 0;
        		
        		//incrementally search the globe until we find a match or search the whole globe//
        		while(!foundMatch && !((i*latSkew > 180) | (i*lonSkew > 360)) ) {
        			double  latSkewFactor = latSkew*i; //increase the search area until we find one 
        			double  lonSkewFactor = lonSkew*i; //increase the search area until we find one 
        			
        			Iterator iter = dataList.iterator(); 
            		ArrayList e = new ArrayList();
            		Atom[] a = new Atom[3];
        			//search each country for a match//
	        		while(iter.hasNext() && (foundCountries<numCountriesToGet)) {	
	        			e = (ArrayList)iter.next();
	        			if(e.size() > 7) {     				
	        				double lat = Double.parseDouble((String)e.get(latIndex));
	        				double lon = Double.parseDouble((String)e.get(lonIndex));
	        					
	        				if(lat < (latIn+latSkewFactor) && lat > (latIn-latSkewFactor)) {
	        					if(lon < (lonIn+lonSkewFactor) && lon > (lonIn-lonSkewFactor)) {
	        						foundMatch = true;
	        						foundCountries++;
		        					//append the messageBox//
		        	        			a = new Atom[3];
		        	        			a[0] = Atom.newAtom(new String("latLonSearchResults")); 
		        	  				a[1] = Atom.newAtom(new String("append"));
		        	  				a[2] = Atom.newAtom((String)e.get(0)); 
		        	  				outlet(2, a );
	        					}
	        				} 
	        			}
	        			
	        		}
	        		i++; //increment the index and continue
        		}	
        		
        		Atom[] aDone = new Atom[2];
        		aDone[0] = Atom.newAtom("latLonSearchResults");
        		aDone[1] = Atom.newAtom("done");
    			outlet(2, aDone);
        }
        
        //***** output matching country data *****//
        public void getCountryData(Atom[] args) {
        		
        		String [] cStringArray = Atom.toString(args);
        		String c = new String("");
        		
        		
        		
        		Atom[] aClear = new Atom[2];
    			aClear[0] = Atom.newAtom("countryData");
    			aClear[1] = Atom.newAtom("set");
    			outlet(2, aClear);
    			
    			//loop through all the countries we found//
    			for(int index=0; index<cStringArray.length && index<numCountriesToGet; index++) {
    				Iterator iter = dataList.iterator(); 
            		ArrayList e = null;
    				c = new String(cStringArray[index]);
    				
    				//loop and test for a match with all the countries
    				while(iter.hasNext()) {	
    					e = (ArrayList)iter.next();
    					if(e.size()<0)
    						continue;
    					
    					if(c.equals((String)e.get(0)) ) {
    						//post("Country Found: " + (String)e.get(0));
    						Atom [] a = new Atom[3];
    						for(int i=0; i<e.size(); i++) {
    							a[0] = Atom.newAtom(new String("countryData")); 
    							a[1] = Atom.newAtom(new String("append"));
    							a[2] = Atom.newAtom((String)e.get(i)); 
    							outlet(2, a );
    						}
    						break; //exit while loop
    					}
    				}
    				
    			}
        	
        }
        
        //**** retrieve indivudal data element *****//
        public void getDataElement(Atom[] args) {
        		String [] cStringArray = Atom.toString(args);
        		
        		ArrayList elementList = new ArrayList();
        		// set elementList according to normalize flag
        		if(cStringArray[0].equals("0"))
        			elementList = new ArrayList(dataList);
        		else if(cStringArray[0].equals("1"))
        			elementList = new ArrayList(dataListNormalized);
        		
        		//post("String:" + cStringArray[0] + ":" + cStringArray[1] );
        		if(cStringArray[1].equals("s") || cStringArray[2].equals("s")) 
        			return;
//        		for(int j=0; j<cStringArray.length;j++)
//        			post("Incoming String :" + cStringArray[j]);	
        		
        		int matchesFound = 0;
			
			//loop through all the requested countries, start at one for norm flag //
        		for(int i=1; i<cStringArray.length && matchesFound<numCountriesToGet; i=i+2) {
        			String countryString = cStringArray[i];
        			String elementString = cStringArray[i+1];
//        			post("Country element i:" + countryString + " " + elementString + " " + i);
        			
        			//loop and test for a match with all the countries
        			Iterator iter = elementList.iterator(); 
            		ArrayList e = null;
				while(iter.hasNext() ) {	
					e = (ArrayList)iter.next();
					if(countryString.equals((String)e.get(0))) {
						matchesFound++;
						//post("Matches Found " + matchesFound );
						//output data element
						Atom [] a = new Atom[3];
						a[0] = Atom.newAtom(new String("dataElement")); 
						a[1] = Atom.newAtom((String)e.get(categoryList.indexOf(elementString)) );
						a[2] = Atom.newAtom(Double.parseDouble((String)e.get(categoryList.indexOf(elementString)) ) ); 
						outlet(1, a);
						//post("dataElement: " + e.get(categoryList.indexOf(elementString)));
					}
				}
//				post("Matches Found: " + matchesFound);
        		}
        	
        }
        
        public void normailzeValues() {
        		Iterator iter = dataList.iterator(); 
    			ArrayList e = null;
    			ArrayList cList = new ArrayList();
 
    			double maxVal = 0.;
    			double normVal = 0.;
    			
    			while(iter.hasNext() ) {	
    				cList = new ArrayList();
    				e = (ArrayList)iter.next();
    				for(int i=0; i<e.size(); i++) {
    					String sVal = (String)e.get(i);
    					maxVal = getMaximumValue(categories[i]);
    					// if getMaximumValue return valid value
    					if(maxVal > -1) {
    						normVal = Double.parseDouble(sVal) / maxVal; 
    						cList.add(i, new String("" + normVal));
    						//post("Normalize: " + categories[i] + " " + s + " = " + normVal);
    					}
    					else {
    						cList.add(i, sVal);
    					}
    				}	
    				dataListNormalized.add(cList);
    			}
        }
        
        public void getMaximumValueOut(String cat) {
	        	double newVal = 0;
	        	String country = new String("");
	        	double max = getMaximumValue(cat);
	        	
	        	Iterator iter = dataList.iterator(); 
	        	ArrayList eList = null;
	        	while(iter.hasNext() ) {	
	        		eList = (ArrayList)iter.next();
	        		newVal = Double.parseDouble((String)eList.get(categoryList.indexOf(cat)));
	        		if(newVal == max) {
	        			country = (String)eList.get(0);
	        			post("Max value for " + cat + " is: " + max + " from " + (String)eList.get(0) );
	        		}
	        	}
	        	Atom [] a = new Atom[3];
	        	a[0] = Atom.newAtom(new String("maxValue")); 
	        	a[1] = Atom.newAtom(max); 
	        	a[2] = Atom.newAtom(country); 
	        	outlet(1, a);
        }
        
        public double getMaximumValue(String cat) {
        		double max = Double.NEGATIVE_INFINITY;
        		double newVal = 0.0;
        		Iterator iter = dataList.iterator(); 
        		ArrayList eList = null;

        		//get cat index    		
			while(iter.hasNext() ) {	
				eList = (ArrayList)iter.next();
				try{
					newVal = Double.parseDouble((String)eList.get(categoryList.indexOf(cat)));
				}
				catch (NumberFormatException e) 
	             {  
					return -1;	               
	             }  // End of catch NumberFormatException 	
				max = Math.max(max, newVal );	
			}        		
        		return max;
        }
        
        public void getMinimumValueOut(String cat) {
        	double newVal = 0;
        	String country = new String("");
        	double min = getMinimumValue(cat);
        	
        	Iterator iter = dataList.iterator(); 
        	ArrayList eList = null;
        	while(iter.hasNext() ) {	
        		eList = (ArrayList)iter.next();
        		newVal = Double.parseDouble((String)eList.get(categoryList.indexOf(cat)));
        		if(newVal == min) {
        			country = (String)eList.get(0);
        			post("Min value for " + cat + " is: " + min + " from " + (String)eList.get(0) );
        		}
        	}
        	Atom [] a = new Atom[3];
        	a[0] = Atom.newAtom(new String("minValue")); 
        	a[1] = Atom.newAtom(min); 
        	a[2] = Atom.newAtom(country); 
        	outlet(1, a);
    }
        
        public double getMinimumValue(String cat) {
	        	double min = Double.POSITIVE_INFINITY;
	        	double newVal = 0.0;
	        	Iterator iter = dataList.iterator(); 
	        	ArrayList eList = null;
	        	
	        	//get cat index    		
	        	while(iter.hasNext() ) {	
	        		eList = (ArrayList)iter.next();
	        		try{
	        			newVal = Double.parseDouble((String)eList.get(categoryList.indexOf(cat)));
	        		}
	        		catch (NumberFormatException e) 
	        		{  
	        			return -1;	               
	        		}  // End of catch NumberFormatException 	
	        		if(newVal != -1)
	        			min = Math.min(min, newVal );	
	        	}        		
	        	return min;
        }
        
        //***** set numCountriesToGet *****//
        public void setNumCountriesToGet(int num){
        		numCountriesToGet = num;
        }
        
        //***** set numCountriesToGet *****//
        public void setLatSkew(double num){
        		latSkew = num;
        }
        
        //***** set numCountriesToGet *****//
        public void setLonSkew(double num){
        		lonSkew = num;
        }
        
        //***** print with many linebreaks to avoid max crashing from huge string in window *****//
        public void println() {
        		post("\n*************************************************");
        		post("*************************************************\n");
        		
			post("dataList: ");
        		Iterator iter = dataList.iterator(); 
        		ArrayList e = null;
        		while(iter.hasNext()) {	
        			e = (ArrayList)iter.next();
        			post(e.toString());
        			
        			
        		}
        }
        
        public void printlnNormalized() {
    		post("\n*************************************************");
    		post("*************************************************\n");
    		
		post("dataListNormalized: ");
    		Iterator iter = dataListNormalized.iterator(); 
    		ArrayList e = null;
    		while(iter.hasNext()) {	
    			e = (ArrayList)iter.next();
    			post(e.toString());
    			
    			
    		}
    }

}
        
